@@ -57,7 +57,7 @@ math_1(PyObject *args, double (*func) (double), char *argsfmt)
 	PyFPE_START_PROTECT("in math_1", return 0)
 	x = (*func)(x);
 	PyFPE_END_PROTECT(x)
-	Py_SET_ERANGE_IF_OVERFLOW(x);
+	Py_SET_ERRNO_ON_MATH_ERROR(x);
 	if (errno && is_error(x))
 		return NULL;
 	else
@@ -74,7 +74,7 @@ math_2(PyObject *args, double (*func) (double, double), char *argsfmt)
 	PyFPE_START_PROTECT("in math_2", return 0)
 	x = (*func)(x, y);
 	PyFPE_END_PROTECT(x)
-	Py_SET_ERANGE_IF_OVERFLOW(x);
+	Py_SET_ERRNO_ON_MATH_ERROR(x);
 	if (errno && is_error(x))
 		return NULL;
 	else
@@ -85,13 +85,13 @@ math_2(PyObject *args, double (*func) (double, double), char *argsfmt)
 	static PyObject * math_##funcname(PyObject *self, PyObject *args) { \
 		return math_1(args, func, "d:" #funcname); \
 	}\
-        static char math_##funcname##_doc [] = docstring;
+        PyDoc_STRVAR(math_##funcname##_doc, docstring);
 
 #define FUNC2(funcname, func, docstring) \
 	static PyObject * math_##funcname(PyObject *self, PyObject *args) { \
 		return math_2(args, func, "dd:" #funcname); \
 	}\
-        static char math_##funcname##_doc [] = docstring;
+        PyDoc_STRVAR(math_##funcname##_doc, docstring);
 
 FUNC1(acos, acos,
       "acos(x)\n\nReturn the arc cosine (measured in radians) of x.")
@@ -121,13 +121,8 @@ FUNC2(fmod, fmod,
       "  x % y may differ.")
 FUNC2(hypot, hypot,
       "hypot(x,y)\n\nReturn the Euclidean distance, sqrt(x*x + y*y).")
-#ifdef MPW_3_1 /* This hack is needed for MPW 3.1 but not for 3.2 ... */
-FUNC2(pow, power,
-      "pow(x,y)\n\nReturn x**y (x to the power of y).")
-#else
 FUNC2(pow, pow,
       "pow(x,y)\n\nReturn x**y (x to the power of y).")
-#endif
 FUNC1(sin, sin,
       "sin(x)\n\nReturn the sine of x (measured in radians).")
 FUNC1(sinh, sinh,
@@ -148,19 +143,19 @@ math_frexp(PyObject *self, PyObject *args)
 		return NULL;
 	errno = 0;
 	x = frexp(x, &i);
-	Py_SET_ERANGE_IF_OVERFLOW(x);
+	Py_SET_ERRNO_ON_MATH_ERROR(x);
 	if (errno && is_error(x))
 		return NULL;
 	else
 		return Py_BuildValue("(di)", x, i);
 }
 
-static char math_frexp_doc [] =
+PyDoc_STRVAR(math_frexp_doc,
 "frexp(x)\n"
 "\n"
 "Return the mantissa and exponent of x, as pair (m, e).\n"
 "m is a float and e is an int, such that x = m * 2.**e.\n"
-"If x is 0, m and e are both 0.  Else 0.5 <= abs(m) < 1.0.";
+"If x is 0, m and e are both 0.  Else 0.5 <= abs(m) < 1.0.");
 
 static PyObject *
 math_ldexp(PyObject *self, PyObject *args)
@@ -173,15 +168,15 @@ math_ldexp(PyObject *self, PyObject *args)
 	PyFPE_START_PROTECT("ldexp", return 0)
 	x = ldexp(x, exp);
 	PyFPE_END_PROTECT(x)
-	Py_SET_ERANGE_IF_OVERFLOW(x);
+	Py_SET_ERRNO_ON_MATH_ERROR(x);
 	if (errno && is_error(x))
 		return NULL;
 	else
 		return PyFloat_FromDouble(x);
 }
 
-static char math_ldexp_doc [] =
-"ldexp(x, i) -> x * (2**i)";
+PyDoc_STRVAR(math_ldexp_doc,
+"ldexp(x, i) -> x * (2**i)");
 
 static PyObject *
 math_modf(PyObject *self, PyObject *args)
@@ -190,27 +185,19 @@ math_modf(PyObject *self, PyObject *args)
 	if (! PyArg_ParseTuple(args, "d:modf", &x))
 		return NULL;
 	errno = 0;
-#ifdef MPW /* MPW C modf expects pointer to extended as second argument */
-        {
-		extended e;
-		x = modf(x, &e);
-		y = e;
-        }
-#else
 	x = modf(x, &y);
-#endif
-	Py_SET_ERANGE_IF_OVERFLOW(x);
+	Py_SET_ERRNO_ON_MATH_ERROR(x);
 	if (errno && is_error(x))
 		return NULL;
 	else
 		return Py_BuildValue("(dd)", x, y);
 }
 
-static char math_modf_doc [] =
+PyDoc_STRVAR(math_modf_doc,
 "modf(x)\n"
 "\n"
 "Return the fractional and integer parts of x.  Both results carry the sign\n"
-"of x.  The integer part is returned as a real.";
+"of x.  The integer part is returned as a real.");
 
 /* A decent logarithm is easy to compute even for huge longs, but libm can't
    do that by itself -- loghelper can.  func is log or log10, and name is
@@ -221,18 +208,8 @@ static char math_modf_doc [] =
 */
 
 static PyObject*
-loghelper(PyObject* args, double (*func)(double), char *name)
+loghelper(PyObject* args, double (*func)(double), char *format, PyObject *arg)
 {
-	PyObject *arg;
-	char format[16];
-
-	/* See whether this is a long. */
-	format[0] = 'O';
-	format[1] = ':';
-	strcpy(format + 2, name);
-	if (! PyArg_ParseTuple(args, format, &arg))
-		return NULL;
-
 	/* If it is long, do it ourselves. */
 	if (PyLong_Check(arg)) {
 		double x;
@@ -252,28 +229,91 @@ loghelper(PyObject* args, double (*func)(double), char *name)
 	}
 
 	/* Else let libm handle it by itself. */
-	format[0] = 'd';
 	return math_1(args, func, format);
 }
 
 static PyObject *
 math_log(PyObject *self, PyObject *args)
 {
-	return loghelper(args, log, "log");
+	PyObject *arg;
+	PyObject *base = NULL;
+	PyObject *num, *den;
+	PyObject *ans;
+	PyObject *newargs;
+
+	if (!PyArg_UnpackTuple(args, "log", 1, 2, &arg, &base))
+		return NULL;
+	if (base == NULL)
+		return loghelper(args, log, "d:log", arg);
+
+	newargs = PyTuple_Pack(1, arg);
+	if (newargs == NULL)
+		return NULL;
+	num = loghelper(newargs, log, "d:log", arg);
+	Py_DECREF(newargs);
+	if (num == NULL)
+		return NULL;
+
+	newargs = PyTuple_Pack(1, base);
+	if (newargs == NULL) {
+		Py_DECREF(num);
+		return NULL;
+	}
+	den = loghelper(newargs, log, "d:log", base);
+	Py_DECREF(newargs);
+	if (den == NULL) {
+		Py_DECREF(num);
+		return NULL;
+	}
+
+	ans = PyNumber_Divide(num, den);
+	Py_DECREF(num);
+	Py_DECREF(den);
+	return ans;
 }
 
-static char math_log_doc[] =
-"log(x) -> the natural logarithm (base e) of x.";
+PyDoc_STRVAR(math_log_doc,
+"log(x[, base]) -> the logarithm of x to the given base.\n\
+If the base not specified, returns the natural logarithm (base e) of x.");
 
 static PyObject *
 math_log10(PyObject *self, PyObject *args)
 {
-	return loghelper(args, log10, "log10");
+	PyObject *arg;
+
+	if (!PyArg_UnpackTuple(args, "log10", 1, 1, &arg))
+		return NULL;
+	return loghelper(args, log10, "d:log10", arg);
 }
 
-static char math_log10_doc[] =
-"log10(x) -> the base 10 logarithm of x.";
+PyDoc_STRVAR(math_log10_doc,
+"log10(x) -> the base 10 logarithm of x.");
 
+static const double degToRad = 3.141592653589793238462643383 / 180.0;
+
+static PyObject *
+math_degrees(PyObject *self, PyObject *args)
+{
+	double x;
+	if (! PyArg_ParseTuple(args, "d:degrees", &x))
+		return NULL;
+	return PyFloat_FromDouble(x / degToRad);
+}
+
+PyDoc_STRVAR(math_degrees_doc,
+"degrees(x) -> converts angle x from radians to degrees");
+
+static PyObject *
+math_radians(PyObject *self, PyObject *args)
+{
+	double x;
+	if (! PyArg_ParseTuple(args, "d:radians", &x))
+		return NULL;
+	return PyFloat_FromDouble(x * degToRad);
+}
+
+PyDoc_STRVAR(math_radians_doc,
+"radians(x) -> converts angle x from degrees to radians");
 
 static PyMethodDef math_methods[] = {
 	{"acos",	math_acos,	METH_VARARGS,	math_acos_doc},
@@ -283,6 +323,7 @@ static PyMethodDef math_methods[] = {
 	{"ceil",	math_ceil,	METH_VARARGS,	math_ceil_doc},
 	{"cos",		math_cos,	METH_VARARGS,	math_cos_doc},
 	{"cosh",	math_cosh,	METH_VARARGS,	math_cosh_doc},
+	{"degrees",	math_degrees,	METH_VARARGS,	math_degrees_doc},
 	{"exp",		math_exp,	METH_VARARGS,	math_exp_doc},
 	{"fabs",	math_fabs,	METH_VARARGS,	math_fabs_doc},
 	{"floor",	math_floor,	METH_VARARGS,	math_floor_doc},
@@ -294,6 +335,7 @@ static PyMethodDef math_methods[] = {
 	{"log10",	math_log10,	METH_VARARGS,	math_log10_doc},
 	{"modf",	math_modf,	METH_VARARGS,	math_modf_doc},
 	{"pow",		math_pow,	METH_VARARGS,	math_pow_doc},
+	{"radians",	math_radians,	METH_VARARGS,	math_radians_doc},
 	{"sin",		math_sin,	METH_VARARGS,	math_sin_doc},
 	{"sinh",	math_sinh,	METH_VARARGS,	math_sinh_doc},
 	{"sqrt",	math_sqrt,	METH_VARARGS,	math_sqrt_doc},
@@ -303,16 +345,18 @@ static PyMethodDef math_methods[] = {
 };
 
 
-static char module_doc [] =
+PyDoc_STRVAR(module_doc,
 "This module is always available.  It provides access to the\n"
-"mathematical functions defined by the C standard.";
+"mathematical functions defined by the C standard.");
 
-DL_EXPORT(void)
+PyMODINIT_FUNC
 initmath(void)
 {
 	PyObject *m, *d, *v;
 
 	m = Py_InitModule3("math", math_methods, module_doc);
+	if (m == NULL)
+		goto finally;
 	d = PyModule_GetDict(m);
 
         if (!(v = PyFloat_FromDouble(atan(1.0) * 4.0)))
