@@ -39,27 +39,20 @@
 #define NOTE(x)
 #endif
 
-#ifdef macintosh
-char *strdup(char *);
-#endif
-
 /*  String constants used to initialize module attributes.
  *
  */
-static char*
-parser_copyright_string
-= "Copyright 1995-1996 by Virginia Polytechnic Institute & State\n\
+static char parser_copyright_string[] =
+"Copyright 1995-1996 by Virginia Polytechnic Institute & State\n\
 University, Blacksburg, Virginia, USA, and Fred L. Drake, Jr., Reston,\n\
 Virginia, USA.  Portions copyright 1991-1995 by Stichting Mathematisch\n\
 Centrum, Amsterdam, The Netherlands.";
 
 
-static char*
-parser_doc_string
-= "This is an interface to Python's internal parser.";
+PyDoc_STRVAR(parser_doc_string,
+"This is an interface to Python's internal parser.");
 
-static char*
-parser_version_string = "0.5";
+static char parser_version_string[] = "0.5";
 
 
 typedef PyObject* (*SeqMaker) (int length);
@@ -92,7 +85,7 @@ node2tuple(node *n,                     /* node to convert               */
         PyObject *v;
         PyObject *w;
 
-        v = mkseq(1 + NCH(n));
+        v = mkseq(1 + NCH(n) + (TYPE(n) == encoding_decl));
         if (v == NULL)
             return (v);
         w = PyInt_FromLong(TYPE(n));
@@ -109,6 +102,9 @@ node2tuple(node *n,                     /* node to convert               */
             }
             (void) addelem(v, i+1, w);
         }
+
+        if (TYPE(n) == encoding_decl)
+            (void) addelem(v, i+1, PyString_FromString(STR(n)));
         return (v);
     }
     else if (ISTERMINAL(TYPE(n))) {
@@ -160,14 +156,9 @@ typedef struct {
 } PyST_Object;
 
 
-staticforward void
-parser_free(PyST_Object *st);
-
-staticforward int
-parser_compare(PyST_Object *left, PyST_Object *right);
-
-staticforward PyObject *
-parser_getattr(PyObject *self, char *name);
+static void parser_free(PyST_Object *st);
+static int parser_compare(PyST_Object *left, PyST_Object *right);
+static PyObject *parser_getattr(PyObject *self, char *name);
 
 
 static
@@ -450,15 +441,15 @@ parser_issuite(PyST_Object *self, PyObject *args, PyObject *kw)
 static PyMethodDef
 parser_methods[] = {
     {"compile",         (PyCFunction)parser_compilest,  PUBLIC_METHOD_TYPE,
-        "Compile this ST object into a code object."},
+        PyDoc_STR("Compile this ST object into a code object.")},
     {"isexpr",          (PyCFunction)parser_isexpr,     PUBLIC_METHOD_TYPE,
-        "Determines if this ST object was created from an expression."},
+        PyDoc_STR("Determines if this ST object was created from an expression.")},
     {"issuite",         (PyCFunction)parser_issuite,    PUBLIC_METHOD_TYPE,
-        "Determines if this ST object was created from a suite."},
+        PyDoc_STR("Determines if this ST object was created from a suite.")},
     {"tolist",          (PyCFunction)parser_st2list,    PUBLIC_METHOD_TYPE,
-        "Creates a list-tree representation of this ST."},
+        PyDoc_STR("Creates a list-tree representation of this ST.")},
     {"totuple",         (PyCFunction)parser_st2tuple,   PUBLIC_METHOD_TYPE,
-        "Creates a tuple-tree representation of this ST."},
+        PyDoc_STR("Creates a tuple-tree representation of this ST.")},
 
     {NULL, NULL, 0, NULL}
 };
@@ -486,7 +477,7 @@ err_string(char *message)
 /*  PyObject* parser_do_parse(PyObject* args, int type)
  *
  *  Internal function to actually execute the parse and return the result if
- *  successful, or set an exception if not.
+ *  successful or set an exception if not.
  *
  */
 static PyObject*
@@ -502,10 +493,8 @@ parser_do_parse(PyObject *args, PyObject *kw, char *argspec, int type)
                                              (type == PyST_EXPR)
                                              ? eval_input : file_input);
 
-        if (n != 0)
-            res = parser_newstobject(n, type);
-        else
-            err_string("could not parse string");
+	if (n)
+	    res = parser_newstobject(n, type);
     }
     return (res);
 }
@@ -553,10 +542,10 @@ parser_suite(PyST_Object *self, PyObject *args, PyObject *kw)
  */
 
 
-staticforward node* build_node_tree(PyObject *tuple);
-staticforward int   validate_expr_tree(node *tree);
-staticforward int   validate_file_input(node *tree);
-
+static node* build_node_tree(PyObject *tuple);
+static int   validate_expr_tree(node *tree);
+static int   validate_file_input(node *tree);
+static int   validate_encoding_decl(node *tree);
 
 /*  PyObject* parser_tuple2st(PyObject* self, PyObject* args)
  *
@@ -609,6 +598,13 @@ parser_tuple2st(PyST_Object *self, PyObject *args, PyObject *kw)
             else
                 PyNode_Free(tree);
         }
+        else if (start_sym == encoding_decl) {
+            /* This looks like an encoding_decl so far. */
+            if (validate_encoding_decl(tree))
+                st = parser_newstobject(tree, PyST_SUITE);
+            else
+                PyNode_Free(tree);
+        }
         else {
             /*  This is a fragment, at best. */
             PyNode_Free(tree);
@@ -694,7 +690,7 @@ build_node_children(PyObject *tuple, node *root, int *line_num)
                         PyErr_Format(parser_error,
                                      "third item in terminal node must be an"
                                      " integer, found %s",
-                                     temp->ob_type->tp_name);
+				     temp->ob_type->tp_name);
                         Py_DECREF(o);
                         Py_DECREF(temp);
                         return 0;
@@ -771,13 +767,28 @@ build_node_tree(PyObject *tuple)
          *  Not efficient, but that can be handled later.
          */
         int line_num = 0;
+        PyObject *encoding = NULL;
 
+        if (num == encoding_decl) {
+            encoding = PySequence_GetItem(tuple, 2);
+            /* tuple isn't borrowed anymore here, need to DECREF */
+            tuple = PySequence_GetSlice(tuple, 0, 2);
+        }
         res = PyNode_New(num);
         if (res != NULL) {
             if (res != build_node_children(tuple, res, &line_num)) {
                 PyNode_Free(res);
                 res = NULL;
             }
+            if (res && encoding) {
+                int len;
+                len = PyString_GET_SIZE(encoding) + 1;
+                res->n_str = (char *)PyMem_MALLOC(len);
+                if (res->n_str != NULL)
+                    (void) memcpy(res->n_str, PyString_AS_STRING(encoding), len);
+                Py_DECREF(encoding);
+                Py_DECREF(tuple);
+            }
         }
     }
     else
@@ -796,7 +807,7 @@ build_node_tree(PyObject *tuple)
 /*
  *  Validation routines used within the validation section:
  */
-staticforward int validate_terminal(node *terminal, int type, char *string);
+static int validate_terminal(node *terminal, int type, char *string);
 
 #define validate_ampersand(ch)  validate_terminal(ch,      AMPER, "&")
 #define validate_circumflex(ch) validate_terminal(ch, CIRCUMFLEX, "^")
@@ -813,6 +824,7 @@ staticforward int validate_terminal(node *terminal, int type, char *string);
 #define validate_vbar(ch)       validate_terminal(ch,       VBAR, "|")
 #define validate_doublestar(ch) validate_terminal(ch, DOUBLESTAR, "**")
 #define validate_dot(ch)        validate_terminal(ch,        DOT, ".")
+#define validate_at(ch)         validate_terminal(ch,         AT, "@")
 #define validate_name(ch, str)  validate_terminal(ch,       NAME, str)
 
 #define VALIDATER(n)    static int validate_##n(node *tree)
@@ -827,6 +839,7 @@ VALIDATER(expr_stmt);           VALIDATER(power);
 VALIDATER(print_stmt);          VALIDATER(del_stmt);
 VALIDATER(return_stmt);         VALIDATER(list_iter);
 VALIDATER(raise_stmt);          VALIDATER(import_stmt);
+VALIDATER(import_name);         VALIDATER(import_from);
 VALIDATER(global_stmt);         VALIDATER(list_if);
 VALIDATER(assert_stmt);         VALIDATER(list_for);
 VALIDATER(exec_stmt);           VALIDATER(compound_stmt);
@@ -844,6 +857,9 @@ VALIDATER(subscriptlist);       VALIDATER(sliceop);
 VALIDATER(exprlist);            VALIDATER(dictmaker);
 VALIDATER(arglist);             VALIDATER(argument);
 VALIDATER(listmaker);           VALIDATER(yield_stmt);
+VALIDATER(testlist1);           VALIDATER(gen_for);
+VALIDATER(gen_iter);            VALIDATER(gen_if);
+VALIDATER(testlist_gexp);
 
 #undef VALIDATER
 
@@ -1056,6 +1072,14 @@ validate_testlist(node *tree)
 }
 
 
+static int
+validate_testlist1(node *tree)
+{
+    return (validate_repeating_list(tree, testlist1,
+                                    validate_test, "testlist1"));
+}
+
+
 static int
 validate_testlist_safe(node *tree)
 {
@@ -1226,6 +1250,21 @@ validate_list_iter(node *tree)
     return res;
 }
 
+/*  gen_iter:  gen_for | gen_if
+ */
+static int
+validate_gen_iter(node *tree)
+{
+    int res = (validate_ntype(tree, gen_iter)
+               && validate_numnodes(tree, 1, "gen_iter"));
+    if (res && TYPE(CHILD(tree, 0)) == gen_for)
+        res = validate_gen_for(CHILD(tree, 0));
+    else
+        res = validate_gen_if(CHILD(tree, 0));
+
+    return res;
+}
+
 /*  list_for:  'for' exprlist 'in' testlist [list_iter]
  */
 static int
@@ -1248,6 +1287,28 @@ validate_list_for(node *tree)
     return res;
 }
 
+/*  gen_for:  'for' exprlist 'in' test [gen_iter]
+ */
+static int
+validate_gen_for(node *tree)
+{
+    int nch = NCH(tree);
+    int res;
+
+    if (nch == 5)
+        res = validate_gen_iter(CHILD(tree, 4));
+    else
+        res = validate_numnodes(tree, 4, "gen_for");
+
+    if (res)
+        res = (validate_name(CHILD(tree, 0), "for")
+               && validate_exprlist(CHILD(tree, 1))
+               && validate_name(CHILD(tree, 2), "in")
+               && validate_test(CHILD(tree, 3)));
+
+    return res;
+}
+
 /*  list_if:  'if' test [list_iter]
  */
 static int
@@ -1268,6 +1329,25 @@ validate_list_if(node *tree)
     return res;
 }
 
+/*  gen_if:  'if' test [gen_iter]
+ */
+static int
+validate_gen_if(node *tree)
+{
+    int nch = NCH(tree);
+    int res;
+
+    if (nch == 3)
+        res = validate_gen_iter(CHILD(tree, 2));
+    else
+        res = validate_numnodes(tree, 2, "gen_if");
+    
+    if (res)
+        res = (validate_name(CHILD(tree, 0), "if")
+               && validate_test(CHILD(tree, 1)));
+
+    return res;
+}
 
 /*  validate_fpdef()
  *
@@ -1635,55 +1715,100 @@ validate_dotted_as_name(node *tree)
 }
 
 
-/*  import_stmt:
- *
- *    'import' dotted_as_name (',' dotted_as_name)*
- *  | 'from' dotted_name 'import' ('*' | import_as_name (',' import_as_name)*)
+/* dotted_as_name (',' dotted_as_name)* */
+static int
+validate_dotted_as_names(node *tree)
+{
+	int nch = NCH(tree);
+	int res = is_odd(nch) && validate_dotted_as_name(CHILD(tree, 0));
+	int i;
+
+	for (i = 1; res && (i < nch); i += 2)
+	    res = (validate_comma(CHILD(tree, i))
+		   && validate_dotted_as_name(CHILD(tree, i + 1)));
+	return (res);
+}
+
+
+/* import_as_name (',' import_as_name)* [','] */
+static int
+validate_import_as_names(node *tree)
+{
+    int nch = NCH(tree);
+    int res = validate_import_as_name(CHILD(tree, 0));
+    int i;
+
+    for (i = 1; res && (i + 1 < nch); i += 2)
+	res = (validate_comma(CHILD(tree, i))
+	       && validate_import_as_name(CHILD(tree, i + 1)));
+    return (res);
+}
+
+
+/* 'import' dotted_as_names */
+static int
+validate_import_name(node *tree)
+{
+	return (validate_ntype(tree, import_name)
+		&& validate_numnodes(tree, 2, "import_name")
+		&& validate_name(CHILD(tree, 0), "import")
+		&& validate_dotted_as_names(CHILD(tree, 1)));
+}
+
+
+/* 'from' dotted_name 'import' ('*' | '(' import_as_names ')' |
+ *     import_as_names
  */
 static int
+validate_import_from(node *tree)
+{
+	int nch = NCH(tree);
+	int res = validate_ntype(tree, import_from)
+		  && (nch >= 4)
+		  && validate_name(CHILD(tree, 0), "from")
+		  && validate_dotted_name(CHILD(tree, 1))
+		  && validate_name(CHILD(tree, 2), "import");
+
+	if (res && TYPE(CHILD(tree, 3)) == LPAR)
+	    res = ((nch == 6)
+		   && validate_lparen(CHILD(tree, 3))
+		   && validate_import_as_names(CHILD(tree, 4))
+		   && validate_rparen(CHILD(tree, 5)));
+	else if (res && TYPE(CHILD(tree, 3)) != STAR)
+	    res = validate_import_as_names(CHILD(tree, 3));
+	return (res);
+}
+
+
+/* import_stmt: import_name | import_from */
+static int
 validate_import_stmt(node *tree)
 {
     int nch = NCH(tree);
-    int res = (validate_ntype(tree, import_stmt)
-               && (nch >= 2) && is_even(nch)
-               && validate_ntype(CHILD(tree, 0), NAME));
+    int res = validate_numnodes(tree, 1, "import_stmt");
 
-    if (res && (strcmp(STR(CHILD(tree, 0)), "import") == 0)) {
-        int j;
+    if (res) {
+	int ntype = TYPE(CHILD(tree, 0));
 
-        res = validate_dotted_as_name(CHILD(tree, 1));
-        for (j = 2; res && (j < nch); j += 2)
-            res = (validate_comma(CHILD(tree, j))
-                   && validate_dotted_as_name(CHILD(tree, j + 1)));
-    }
-    else if (res && (res = validate_name(CHILD(tree, 0), "from"))) {
-        res = ((nch >= 4) && is_even(nch)
-               && validate_dotted_name(CHILD(tree, 1))
-               && validate_name(CHILD(tree, 2), "import"));
-        if (nch == 4) {
-            if (TYPE(CHILD(tree, 3)) == import_as_name)
-                res = validate_import_as_name(CHILD(tree, 3));
-            else
-                res = validate_star(CHILD(tree, 3));
-        }
+	if (ntype == import_name || ntype == import_from)
+            res = validate_node(CHILD(tree, 0));
         else {
-            /*  'from' dotted_name 'import' import_as_name
-             *      (',' import_as_name)+
-             */
-            int j;
-            res = validate_import_as_name(CHILD(tree, 3));
-            for (j = 4; res && (j < nch); j += 2)
-                res = (validate_comma(CHILD(tree, j))
-                       && validate_import_as_name(CHILD(tree, j + 1)));
+            res = 0;
+            err_string("illegal import_stmt child type");
         }
     }
-    else
+    else if (nch == 1) {
         res = 0;
-
+        PyErr_Format(parser_error,
+                     "Unrecognized child node of import_stmt: %d.",
+                     TYPE(CHILD(tree, 0)));
+    }
     return (res);
 }
 
 
+
+
 static int
 validate_global_stmt(node *tree)
 {
@@ -2167,7 +2292,7 @@ validate_atom(node *tree)
                    && (validate_rparen(CHILD(tree, nch - 1))));
 
             if (res && (nch == 3))
-                res = validate_testlist(CHILD(tree, 1));
+                res = validate_testlist_gexp(CHILD(tree, 1));
             break;
           case LSQB:
             if (nch == 2)
@@ -2189,7 +2314,7 @@ validate_atom(node *tree)
             break;
           case BACKQUOTE:
             res = ((nch == 3)
-                   && validate_testlist(CHILD(tree, 1))
+                   && validate_testlist1(CHILD(tree, 1))
                    && validate_ntype(CHILD(tree, 2), BACKQUOTE));
             break;
           case NAME:
@@ -2224,7 +2349,7 @@ validate_listmaker(node *tree)
         ok = validate_test(CHILD(tree, 0));
 
     /*
-     *  list_iter | (',' test)* [',']
+     *  list_for | (',' test)* [',']
      */
     if (nch == 2 && TYPE(CHILD(tree, 1)) == list_for)
         ok = validate_list_for(CHILD(tree, 1));
@@ -2246,21 +2371,106 @@ validate_listmaker(node *tree)
     return ok;
 }
 
+/*  testlist_gexp:
+ *    test ( gen_for | (',' test)* [','] )
+ */
+static int
+validate_testlist_gexp(node *tree)
+{
+    int nch = NCH(tree);
+    int ok = nch;
+
+    if (nch == 0)
+        err_string("missing child nodes of testlist_gexp");
+    else {
+        ok = validate_test(CHILD(tree, 0));
+    }
+
+    /*
+     *  gen_for | (',' test)* [',']
+     */
+    if (nch == 2 && TYPE(CHILD(tree, 1)) == gen_for)
+        ok = validate_gen_for(CHILD(tree, 1));
+    else {
+        /*  (',' test)* [',']  */
+        int i = 1;
+        while (ok && nch - i >= 2) {
+            ok = (validate_comma(CHILD(tree, i))
+                  && validate_test(CHILD(tree, i+1)));
+            i += 2;
+        }
+        if (ok && i == nch-1)
+            ok = validate_comma(CHILD(tree, i));
+        else if (i != nch) {
+            ok = 0;
+            err_string("illegal trailing nodes for testlist_gexp");
+        }
+    }
+    return ok;
+}
+
+/*  decorator:
+ *    '@' dotted_name [ '(' [arglist] ')' ] NEWLINE
+ */
+static int
+validate_decorator(node *tree)
+{
+    int ok;
+    int nch = NCH(tree);
+    ok = (validate_ntype(tree, decorator) &&
+	  (nch == 3 || nch == 5 || nch == 6) &&
+	  validate_at(CHILD(tree, 0)) &&
+	  validate_dotted_name(CHILD(tree, 1)) &&
+	  validate_newline(RCHILD(tree, -1)));
+
+    if (ok && nch != 3) {
+	ok = (validate_lparen(CHILD(tree, 2)) &&
+	      validate_rparen(RCHILD(tree, -2)));
+
+	if (ok && nch == 6)
+	    ok = validate_arglist(CHILD(tree, 3));
+    }
+
+    return ok;
+}
+
+/*  decorators:
+ *    decorator+
+ */
+static int
+validate_decorators(node *tree)
+{
+    int i, nch, ok; 
+    nch = NCH(tree);
+    ok = validate_ntype(tree, decorators) && nch >= 1;
+
+    for (i = 0; ok && i < nch; ++i)
+	ok = validate_decorator(CHILD(tree, i));
+
+    return ok;
+}
 
 /*  funcdef:
- *      'def' NAME parameters ':' suite
- *
+ *      
+ *            -6   -5    -4         -3  -2 -1
+ *  [decorators] 'def' NAME parameters ':' suite
  */
 static int
 validate_funcdef(node *tree)
 {
-    return (validate_ntype(tree, funcdef)
-            && validate_numnodes(tree, 5, "funcdef")
-            && validate_name(CHILD(tree, 0), "def")
-            && validate_ntype(CHILD(tree, 1), NAME)
-            && validate_colon(CHILD(tree, 3))
-            && validate_parameters(CHILD(tree, 2))
-            && validate_suite(CHILD(tree, 4)));
+    int nch = NCH(tree);
+    int ok = (validate_ntype(tree, funcdef)
+	       && ((nch == 5) || (nch == 6))
+	       && validate_name(RCHILD(tree, -5), "def")
+	       && validate_ntype(RCHILD(tree, -4), NAME)
+	       && validate_colon(RCHILD(tree, -2))
+	       && validate_parameters(RCHILD(tree, -3))
+	       && validate_suite(RCHILD(tree, -1)));
+
+    if (ok && (nch == 6))
+	ok = validate_decorators(CHILD(tree, 0));
+
+    return ok;
 }
 
 
@@ -2298,6 +2508,18 @@ validate_arglist(node *tree)
         /* raise the right error from having an invalid number of children */
         return validate_numnodes(tree, nch + 1, "arglist");
 
+    if (nch > 1) {
+        for (i=0; i<nch; i++) {
+            if (TYPE(CHILD(tree, i)) == argument) {
+                node *ch = CHILD(tree, i);
+                if (NCH(ch) == 2 && TYPE(CHILD(ch, 1)) == gen_for) {
+                    err_string("need '(', ')' for generator expression");
+                    return 0;
+                }
+            }
+        }
+    }
+
     while (ok && nch-i >= 2) {
         /* skip leading (argument ',') */
         ok = (validate_argument(CHILD(tree, i))
@@ -2357,17 +2579,19 @@ validate_arglist(node *tree)
 
 /*  argument:
  *
- *  [test '='] test
+ *  [test '='] test [gen_for]
  */
 static int
 validate_argument(node *tree)
 {
     int nch = NCH(tree);
     int res = (validate_ntype(tree, argument)
-               && ((nch == 1) || (nch == 3))
+               && ((nch == 1) || (nch == 2) || (nch == 3))
                && validate_test(CHILD(tree, 0)));
 
-    if (res && (nch == 3))
+    if (res && (nch == 2))
+        res = validate_gen_for(CHILD(tree, 1));
+    else if (res && (nch == 3))
         res = (validate_equal(CHILD(tree, 1))
                && validate_test(CHILD(tree, 2)));
 
@@ -2645,6 +2869,12 @@ validate_node(node *tree)
           case import_stmt:
             res = validate_import_stmt(tree);
             break;
+	  case import_name:
+	    res = validate_import_name(tree);
+	    break;
+	  case import_from:
+	    res = validate_import_from(tree);
+	    break;
           case global_stmt:
             res = validate_global_stmt(tree);
             break;
@@ -2675,6 +2905,9 @@ validate_node(node *tree)
           case testlist:
             res = validate_testlist(tree);
             break;
+          case testlist1:
+            res = validate_testlist1(tree);
+            break;
           case test:
             res = validate_test(tree);
             break;
@@ -2772,6 +3005,18 @@ validate_file_input(node *tree)
     return (res);
 }
 
+static int
+validate_encoding_decl(node *tree)
+{
+    int nch = NCH(tree);
+    int res = ((nch == 1)
+        && validate_file_input(CHILD(tree, 0)));
+
+    if (!res && !PyErr_Occurred())
+        err_string("Error Parsing encoding_decl");
+
+    return res;
+}
 
 static PyObject*
 pickle_constructor = NULL;
@@ -2816,60 +3061,70 @@ parser__pickler(PyObject *self, PyObject *args)
  */
 static PyMethodDef parser_functions[] =  {
     {"ast2tuple",       (PyCFunction)parser_st2tuple,  PUBLIC_METHOD_TYPE,
-        "Creates a tuple-tree representation of an ST."},
+        PyDoc_STR("Creates a tuple-tree representation of an ST.")},
     {"ast2list",        (PyCFunction)parser_st2list,   PUBLIC_METHOD_TYPE,
-        "Creates a list-tree representation of an ST."},
+        PyDoc_STR("Creates a list-tree representation of an ST.")},
     {"compileast",      (PyCFunction)parser_compilest, PUBLIC_METHOD_TYPE,
-        "Compiles an ST object into a code object."},
+        PyDoc_STR("Compiles an ST object into a code object.")},
     {"compilest",      (PyCFunction)parser_compilest,  PUBLIC_METHOD_TYPE,
-        "Compiles an ST object into a code object."},
+        PyDoc_STR("Compiles an ST object into a code object.")},
     {"expr",            (PyCFunction)parser_expr,      PUBLIC_METHOD_TYPE,
-        "Creates an ST object from an expression."},
+        PyDoc_STR("Creates an ST object from an expression.")},
     {"isexpr",          (PyCFunction)parser_isexpr,    PUBLIC_METHOD_TYPE,
-        "Determines if an ST object was created from an expression."},
+        PyDoc_STR("Determines if an ST object was created from an expression.")},
     {"issuite",         (PyCFunction)parser_issuite,   PUBLIC_METHOD_TYPE,
-        "Determines if an ST object was created from a suite."},
+        PyDoc_STR("Determines if an ST object was created from a suite.")},
     {"suite",           (PyCFunction)parser_suite,     PUBLIC_METHOD_TYPE,
-        "Creates an ST object from a suite."},
+        PyDoc_STR("Creates an ST object from a suite.")},
     {"sequence2ast",    (PyCFunction)parser_tuple2st,  PUBLIC_METHOD_TYPE,
-        "Creates an ST object from a tree representation."},
+        PyDoc_STR("Creates an ST object from a tree representation.")},
     {"sequence2st",     (PyCFunction)parser_tuple2st,  PUBLIC_METHOD_TYPE,
-        "Creates an ST object from a tree representation."},
+        PyDoc_STR("Creates an ST object from a tree representation.")},
     {"st2tuple",        (PyCFunction)parser_st2tuple,  PUBLIC_METHOD_TYPE,
-        "Creates a tuple-tree representation of an ST."},
+        PyDoc_STR("Creates a tuple-tree representation of an ST.")},
     {"st2list",         (PyCFunction)parser_st2list,   PUBLIC_METHOD_TYPE,
-        "Creates a list-tree representation of an ST."},
+        PyDoc_STR("Creates a list-tree representation of an ST.")},
     {"tuple2ast",       (PyCFunction)parser_tuple2st,  PUBLIC_METHOD_TYPE,
-        "Creates an ST object from a tree representation."},
+        PyDoc_STR("Creates an ST object from a tree representation.")},
     {"tuple2st",        (PyCFunction)parser_tuple2st,  PUBLIC_METHOD_TYPE,
-        "Creates an ST object from a tree representation."},
+        PyDoc_STR("Creates an ST object from a tree representation.")},
 
     /* private stuff: support pickle module */
     {"_pickler",        (PyCFunction)parser__pickler,  METH_VARARGS,
-        "Returns the pickle magic to allow ST objects to be pickled."},
+        PyDoc_STR("Returns the pickle magic to allow ST objects to be pickled.")},
 
     {NULL, NULL, 0, NULL}
     };
 
 
-DL_EXPORT(void) initparser(void);  /* supply a prototype */
+PyMODINIT_FUNC initparser(void);  /* supply a prototype */
 
-DL_EXPORT(void)
+PyMODINIT_FUNC
 initparser(void)
 {
     PyObject *module, *copyreg;
 
     PyST_Type.ob_type = &PyType_Type;
     module = Py_InitModule("parser", parser_functions);
+    if (module == NULL)
+    	return;
 
     if (parser_error == 0)
         parser_error = PyErr_NewException("parser.ParserError", NULL, NULL);
 
-    if ((parser_error == 0)
-        || (PyModule_AddObject(module, "ParserError", parser_error) != 0)) {
+    if (parser_error == 0)
         /* caller will check PyErr_Occurred() */
         return;
-    }
+    /* CAUTION:  The code next used to skip bumping the refcount on
+     * parser_error.  That's a disaster if initparser() gets called more
+     * than once.  By incref'ing, we ensure that each module dict that
+     * gets created owns its reference to the shared parser_error object,
+     * and the file static parser_error vrbl owns a reference too.
+     */
+    Py_INCREF(parser_error);
+    if (PyModule_AddObject(module, "ParserError", parser_error) != 0)
+        return;
+
     Py_INCREF(&PyST_Type);
     PyModule_AddObject(module, "ASTType", (PyObject*)&PyST_Type);
     Py_INCREF(&PyST_Type);
